#include <stdlib.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "em_mog.h"
#include "macros.h"

static void
printf_matrix(const double *m, int ncol, int nrow)
{
	const double *ptr = m;
	int i, j;
	printf("\n");
	for (i = 0; i < nrow; i++) {
		for (j = 0; j < ncol; j++) {
			printf("%f ", *ptr);
			ptr++;
		}
		printf("\n");
	}
}

static double
exp_iir_filter(const double *mean, const double *x, const double *cmx, int dim)
{
	const double *m, *p1, *p2;
	double tmp, d = 0.0;
	int j;
	m  = cmx;
	p1 = mean;
	p2 = x;

	for (j = 0; j < dim; j++) {
		tmp = fabs(*p1 - *p2);
		d += tmp * tmp / (*m + 1);
		p1++;
		p2++;
		m += dim + j + 1;
	}

	return exp(-d);
}

/*
 * em_cholesky_decomp: cholesky decomposition from matrix with dimension 3x3
 */
static int
em_cholesky_decomp_3x3(gsl_matrix *C)
{
	return_if_fail(C->size1 == 3 && C->size2 == 3);
	double *d = C->data;
	double tmp;
	if (*d <= 0)
		goto err;
	/* L[1,1] = sqrt(A[1,1]) */
	*d = sqrt(*d);
	/* L[2,1] = 1 / L[1,1] * (A[2,1])*/
	d[3] /= d[0];
	tmp = d[4] - d[3] * d[3];
	if (tmp <= 0.0)
		goto err;
	d[4] = sqrt(tmp);

	d[6] /= d[0];
	d[7] = (d[7] - d[3] * d[6]) / d[4];
	tmp = d[8] - d[6] * d[6] - d[7] * d[7];
	if (tmp <= 0.0)
		goto err;
	if (tmp <= 0.0)
			goto err;
	d[8] = sqrt(tmp);
	goto quit;

err:
	fprintf(stderr, "em_cholesky_decomp: matrix must be positive definite\n");
	printf_matrix(d, 3, 3);
	return GSL_EDOM;
quit:;
	return 0;
}

static double
normal_distribution(const double *mean, const double *x, const double *cmx, int dim)
{
	return_val_if_fail(mean != NULL, EM_MOG_ERROR);
	return_val_if_fail(x != NULL, EM_MOG_ERROR);
	return_val_if_fail(cmx != NULL, EM_MOG_ERROR);
	return_val_if_fail(dim > 0, EM_MOG_ERROR);
	double *vect;
	double degree = 0.0, ret = 0.0, det = 1.0, log_exp;
	int i, j, res, cmx_size = dim * dim;

	vect = malloc(dim * sizeof(double));
	memset(vect, 0, sizeof(double) * dim);

	gsl_matrix *C = gsl_matrix_alloc(dim, dim);
	memcpy(C->data, cmx, cmx_size * sizeof(double));

	res = gsl_linalg_cholesky_decomp(C);

	return_if_fail(res != GSL_EDOM);

	double *m = C->data;
	/** the solution of linear equations by Gauss for the triangular matrix
	 * |a11   0   0 ... 0|
	 * |a21 a22   0 ... 0|
	 * |.................|
	 * |an1 an2 ...   ann|
	 */
	for (i = 0; i < dim; i++) {
		double tmp = x[i] - mean[i];
		for (j = 0; j < i; j++) {
			tmp -= *m * vect[j];
			m++;
		}
		vect[i] = tmp / *m;
		det = *m * det;
		degree += vect[i] * vect[i];
		m += dim - i;
	}

	log_exp = -0.5 * dim * log(2 * M_PI) - log(det) - 0.5 * degree;

	ret = exp(log_exp);

	gsl_matrix_free(C);
	free(vect);

	return ret;
}

static double
log_exp(const double *mean, const double *x, const double *cmx, int dim)
{
	return_val_if_fail(mean != NULL, EM_MOG_ERROR);
	return_val_if_fail(x != NULL, EM_MOG_ERROR);
	return_val_if_fail(cmx != NULL, EM_MOG_ERROR);
	return_val_if_fail(dim > 0, EM_MOG_ERROR);
	double *vect;
	double degree = 0.0, det = 1.0, log_exp;
	int i, j, res, cmx_size = dim * dim;

	vect = malloc(dim * sizeof(double));
	memset(vect, 0, sizeof(double) * dim);

	gsl_matrix *C = gsl_matrix_alloc(dim, dim);

	memcpy(C->data, cmx, cmx_size * sizeof(double));

	res = gsl_linalg_cholesky_decomp(C);

	return_if_fail(res != GSL_EDOM);

	double *m = C->data;
	/** the solution of linear equations by Gauss for the triangular matrix
	 * |a11   0   0 ... 0|
	 * |a21 a22   0 ... 0|
	 * |.................|
	 * |an1 an2 ...   ann|
	 */
	for (i = 0; i < dim; i++) {
		double tmp = x[i] - mean[i];
		for (j = 0; j < i; j++) {
			tmp -= *m * vect[j];
			m++;
		}
		vect[i] = tmp / *m;
		det = *m * det;
		degree += vect[i] * vect[i];
		m += dim - i;
	}

	log_exp = -0.5 * dim * log(2 * M_PI) - log(det) - 0.5 * degree;

	gsl_matrix_free(C);
	free(vect);

	return log_exp;
}

/** EM algorithm expectation step */
static int
expectation_step(double *probability, const struct em_mog_context *em_ctx,
		         const double *data, int data_size)
{
	return_val_if_fail(probability, EM_MOG_ERROR);
	return_val_if_fail(em_ctx, EM_MOG_ERROR);
	return_val_if_fail(data, EM_MOG_ERROR);
	return_val_if_fail(em_ctx->means, EM_MOG_ERROR);

	double denum, *p1, *p2, *m, *cov_mtx;
	const double *x;
	int i, j, n, dim, cov_size;
	n = em_ctx->ngauss;
	dim      = em_ctx->dim;
	cov_size = dim * dim;
	p1 = p2 = probability;
	x = data;

	for (i = 0; i < data_size; i++) {
		denum = 0.0;
		m       = em_ctx->means;
		cov_mtx = em_ctx->cov_matrix;

		for (j = 0; j < em_ctx->ngauss; j++) {
			*p1 = em_ctx->weight[j] * normal_distribution(m, x, cov_mtx, dim);
			denum += *p1;
			m += dim;
			cov_mtx += cov_size;
			p1++;
		}

		for (j = 0; j < em_ctx->ngauss; j++) {
			if (denum > 0)
				*p2 /= denum;
			p2++;
		}
		x += dim;
	}

	return EM_MOG_SUCCESS;
}

static void
add_vector_sqr_to_matrix(double *matrix, const double *vector, double mult, int dim)
{
	return_if_fail(matrix != NULL);
	return_if_fail(vector != NULL);

	double *m;
	const double *p1, *p2;
	int i, j;

	m = matrix;
	p1 = vector;
	for (i = 0; i < dim; i++) {
		p2 = vector;
		for (j = 0; j < dim; j++) {
			*m += mult * (*p1) * (*p2);
			if (i == j)
				*m += 0.001;
			m++;
			p2++;
		}
		p1++;
	}
}

static void
add_vector_to_vector(double *dest, const double *sourse, double mult, int dim)
{
	int i;
	double *p1;
	const double *p2;
	p1 = dest;
	p2 = sourse;
	for (i = 0; i < dim; i++) {
		*p1 += *p2 * mult;
		p1++;
		p2++;
	}
}

/** EM algorithm maximization step */
static int
maximization_step(const double *probability, struct em_mog_context *em_ctx,
		          const double *data, int data_size)
{
	return_val_if_fail(probability, EM_MOG_ERROR);
	return_val_if_fail(em_ctx, EM_MOG_ERROR);
	return_val_if_fail(data, EM_MOG_ERROR);
	return_val_if_fail(em_ctx->means, EM_MOG_ERROR);

	double *w, *m, *cmx, *vector;
	const double *p, *x;
	int i, j, k, dim, cov_size;

	w = em_ctx->weight;
	m = em_ctx->means;
	cmx = em_ctx->cov_matrix;

	dim = em_ctx->dim;
	cov_size = dim * dim;
	vector = malloc(dim * sizeof(double));

	memset(w, 0, em_ctx->ngauss * sizeof(double));
	memset(m, 0, em_ctx->ngauss * dim * sizeof(double));
	memset(cmx, 0, em_ctx->ngauss * cov_size * sizeof(double));

	for (i = 0; i < em_ctx->ngauss; i++) {
		p = probability + i;
		for (j = 0; j < data_size; j++) {
			*w += *p;
			p += em_ctx->ngauss;
		}

		p = probability + i;
		for (j = 0, x = data; j < data_size; j++) {
			double mult = *p / *w;
			add_vector_to_vector(m, x, mult, dim);
			p += em_ctx->ngauss;
			x += dim;
		}

		p = probability + i;
		for (j = 0, x = data; j < data_size; j++) {
			for (k = 0; k < dim; k++)
				vector[k] = x[k] - m[k];
			double mult = *p / *w;
			add_vector_sqr_to_matrix(cmx, vector, mult, dim);
			p += em_ctx->ngauss;
			x += dim;
		}
		*w /= data_size;
		w++;
		m += dim;
		cmx += cov_size;
	}

	free(vector);

	return EM_MOG_SUCCESS;
}

static double
em_likelihood(struct em_mog_context *em_ctx, const double *data, int data_size)
{
	double *log_vect;;
	const double *x, *w, *m, *cmx;

	double sum_exp, loglike = 0;
	int i, j, dim, cov_size;
	dim = em_ctx->dim;
	cov_size = dim * dim;
	em_ctx->min_loglike = 1.0;

	log_vect = malloc(em_ctx->ngauss * sizeof(double));

	x = data;

	for (i = 0; i < data_size; i++) {
		w = em_ctx->weight;
		m = em_ctx->means;
		cmx = em_ctx->cov_matrix;
		sum_exp = 0;

		double tmp = log(*w) + log_exp(m, x, cmx, dim);
		log_vect[0] = tmp;
		w++;
		m   += dim;
		cmx += cov_size;

		for (j = 1; j < em_ctx->ngauss; j++) {
			log_vect[j] = log(*w) + log_exp(m, x, cmx, dim);

			w++;
			m   += dim;
			cmx += cov_size;

			if (log_vect[j] > tmp)
				tmp = log_vect[j];
		}

		for (j = 0; j < em_ctx->ngauss; j++)
			sum_exp += exp(log_vect[j] - tmp);

		tmp += sum_exp;

		if (tmp < em_ctx->min_loglike)
			em_ctx->min_loglike = tmp;

		loglike += tmp;
		x += dim;
	}

	free(log_vect);
	return loglike;
}

static double
em_data_probability(const struct em_mog_context *em_ctx, const double *data)
{
	double *log_vect;
	const double *x, *w, *m, *cmx;

	double sum_exp, loglike = 0;
	int j, dim, cov_size;
	dim = em_ctx->dim;
	cov_size = dim * dim;

	log_vect = malloc(em_ctx->ngauss * sizeof(double));

	x = data;
	w = em_ctx->weight;
	m = em_ctx->means;
	cmx = em_ctx->cov_matrix;
	sum_exp = 0;

	double tmp = log(*w) + log_exp(m, x, cmx, dim);
	log_vect[0] = tmp;

	for (j = 1; j < em_ctx->ngauss; j++) {
		w++;
		m   += dim;
		cmx += cov_size;

		log_vect[j] = log(*w) + log_exp(m, x, cmx, dim);

		if (log_vect[j] > tmp)
			tmp = log_vect[j];
	}

	for (j = 0; j < em_ctx->ngauss; j++)
		sum_exp += exp(log_vect[j] - tmp);

	loglike += tmp + sum_exp;


	free(log_vect);
	return loglike;
}

struct em_mog_context *
em_mog_new()
{
	struct em_mog_context *ptr;
	ptr = malloc(sizeof(struct em_mog_context));
	return_if_fail(ptr != NULL);
	ptr->cov_matrix = NULL;
	ptr->means      = NULL;
	ptr->weight     = NULL;
	ptr->dim        = 0;
	ptr->ngauss     = 0;
	return ptr;
}

void
em_mog_free(struct em_mog_context **ctx)
{
	return_if_fail(ctk == NULL);
	return_if_fail((*ctx) == NULL);
	if ((*ctx)->means != NULL)
		free((*ctx)->means);

	free(*ctx);
	*ctx = NULL;
}

void
em_mog_init(struct em_mog_context *ctx)
{
	return_if_fail(ctx != NULL);
	int size, cov_size;
	size     = ctx->dim * ctx->ngauss;
	cov_size = ctx->dim * size;

	return_if_fail(size != 0);

	ctx->means      = malloc((size + ctx->ngauss + cov_size) * sizeof(double));
	return_if_fail(ctx->means != NULL);
	ctx->weight     = ctx->means + size;
	ctx->cov_matrix = ctx->weight + ctx->ngauss;
}

void
em_mog_clear(struct em_mog_context *ctx)
{
	return_if_fail(ctx != NULL);
	if (ctx->means != NULL)
		free(ctx->means);
	ctx->cov_matrix = NULL;
	ctx->means      = NULL;
	ctx->weight     = NULL;
	ctx->dim    = 0;
	ctx->ngauss = 0;
	ctx->epsilon = 0.0;
	ctx->min_loglike = 0.0;
}

/* Set error value for end of iteration process. */
void em_mog_set_error(struct em_mog_context *ctx, double error)
{
	return_val_if_fail(ctx != NULL, -1.0);
	ctx->epsilon = error;
}
/* Get error value for end of iteration process. */
double em_mog_get_error(struct em_mog_context *ctx)
{
	return_val_if_fail(ctx != NULL, -1.0);
	return ctx->epsilon;
}

/* Set number of gaussians in mixture. */
void
em_mog_set_ngaussians(struct em_mog_context *ctx, int ngauss)
{
	return_if_fail(ctx != NULL);
	ctx->ngauss = ngauss;
}

/* Get number of gaussians in mixture. */
int
em_mog_get_ngaussians(struct em_mog_context *ctx)
{
	return_if_fail(ctx != NULL);
	return ctx->ngauss;
}

/* Set fitted data item dimension. */
void
em_mog_set_dimension(struct em_mog_context *ctx, int dim)
{
	return_if_fail(ctx != NULL);
	ctx->dim = dim;
}
/* Get fitted data item dimension. */
int
em_mog_get_dimension(struct em_mog_context *ctx)
{
	return_val_if_fail(ctx != NULL, EM_MOG_ERROR);
	return ctx->dim;
}

/* Set n-th gaussian component weight. */
void
em_mog_set_nth_weight(struct em_mog_context *ctx, int nth, double weight)
{
	return_if_fail(ctx != NULL);
	return_if_fail(ctx->weight != NULL);
	return_if_fail(nth > 0 && nth < ctx->ngauss);

	ctx->weight[nth] = weight;
}

/* Get n-th gaussian component weight. */
double
em_mog_get_nth_weight(struct em_mog_context *ctx, int nth)
{
	return_if_fail(ctx != NULL);
	return_if_fail(ctx->weight != NULL);
	return_if_fail(nth > 0 && nth < ctx->ngauss);

	return ctx->weight[nth];
}

/* Set n-th gaussian mean. */
void
em_mog_set_nth_mean(struct em_mog_context *ctx, int nth, const double *mu)
{
	return_if_fail(ctx != NULL);
	return_if_fail(ctx->means != NULL);
	return_if_fail(nth > 0 && nth < ctx->ngauss);
	return_if_fail(mu != NULL);

	memcpy(&(ctx->means[nth * ctx->dim]), mu, sizeof(double) * ctx->dim);
}
/* Get n-th gaussian mean. */
void
em_mog_get_nth_mean(struct em_mog_context *ctx, int nth, double *mu)
{
	return_if_fail(ctx != NULL);
	return_if_fail(ctx->means != NULL);
	return_if_fail(nth > 0 && nth < ctx->ngauss);
	return_if_fail(mu != NULL);

	memcpy(mu, &(ctx->means[nth * ctx->dim]), sizeof(double) * ctx->dim);
}

/* Set n-th gaussian covariance matrix. */
void
em_mog_set_nth_covariance_matrix(struct em_mog_context *ctx, int nth, const double *cmx)
{
	return_if_fail(ctx != NULL);
	return_if_fail(ctx->cov_matrix != NULL);
	return_if_fail(nth > 0 && nth < ctx->ngauss);
	return_if_fail(cmx != NULL);
	int cov_size = ctx->dim * ctx->dim;

	memcpy(&(ctx->cov_matrix[nth * cov_size]), cmx, sizeof(double) * cov_size);
}
/* Get n-th gaussian covariance matrix. */
void
em_mog_get_nth_covariance_matrix(struct em_mog_context *ctx, int nth, double *cmx)
{
	return_if_fail(ctx != NULL);
	return_if_fail(ctx->cov_matrix != NULL);
	return_if_fail(nth > 0 && nth < ctx->ngauss);
	return_if_fail(cmx != NULL);
	int cov_size = ctx->dim * ctx->dim;

	memcpy(cmx, &(ctx->cov_matrix[nth * cov_size]), sizeof(double) * cov_size);
}

double
em_mog_step(struct em_mog_context *ctx, double *x, int nx)
{
	return_val_if_fail(ctx != NULL, 0.0);
	return_val_if_fail(ctx->means != NULL, 0.0);
	return_val_if_fail(nx > 0, 0.0);
	return_val_if_fail(x != NULL, 0.0);
	double loglike = 0.0;
	int res;

	double *probability = malloc(nx * ctx->ngauss * sizeof(double));
	return_val_if_fail(probability != NULL, loglike);

	res  = expectation_step(probability, ctx, x, nx);
	if (res == EM_MOG_ERROR)
		goto quit;

	res = maximization_step(probability, ctx, x, nx);
	if (res == EM_MOG_ERROR)
		goto quit;

	loglike = em_likelihood(ctx, x, nx);
quit:
	free(probability);
	return loglike;
}

int
em_mog_converge(struct em_mog_context *ctx, double *x, int nx, double epsilon)
{
	return_val_if_fail(ctx != NULL, 0.0);
	return_val_if_fail(ctx->means != NULL, 0.0);
	return_val_if_fail(nx > 0, 0.0);
	return_val_if_fail(x != NULL, 0.0);
	double err, loglike_prev, loglike = 0.0;
	int res;

	double *probability = malloc(nx * ctx->ngauss * sizeof(double));
	return_val_if_fail(probability != NULL, loglike);

	res  = expectation_step(probability, ctx, x, nx);
	if (res == EM_MOG_ERROR)
		goto quit;

	res = maximization_step(probability, ctx, x, nx);
	if (res == EM_MOG_ERROR)
		goto quit;

	loglike_prev = em_likelihood(ctx, x, nx);

	do {
		res  = expectation_step(probability, ctx, x, nx);
		if (res == EM_MOG_ERROR)
			goto quit;

		res = maximization_step(probability, ctx, x, nx);
		if (res == EM_MOG_ERROR)
			goto quit;

		loglike = em_likelihood(ctx, x, nx);
		err = fabs(loglike - loglike_prev);
		loglike_prev = loglike;
	} while (err > epsilon);

/*TODO: убрать отладочный код*/
#ifdef DEBUG
int i, dim, cov_size;

dim = ctx->dim;
cov_size = dim * dim;
for (i = 0; i < ctx->ngauss; i++) {
	printf("%d\n", i + 1);
	printf_matrix(ctx->means + dim * i, dim, 1);
	printf("\n");
	printf("%d\n", i + 1);
	printf_matrix(ctx->cov_matrix + i * cov_size, dim, dim);
	printf("\n");
	printf("%d\n", i + 1);
	printf("%lf ", *(ctx->weight + i));
	printf("\n\n");
}
#endif //DEBUG

quit:
	free(probability);
	return res;
}

int
em_mog_sample_to_match_model(struct em_mog_context *ctx, double *x)
{
	return_val_if_fail(ctx != NULL, EM_MOG_ERROR);
	return_val_if_fail(x != NULL, EM_MOG_ERROR);
	return_val_if_fail(ctx->means != NULL, EM_MOG_ERROR);

	double loglike;
	int ret;

	loglike = em_data_probability(ctx, x);

	ret = (loglike >= ctx->min_loglike) ? EM_MOG_SUCCESS : EM_MOG_ERROR;

	return ret;
}

void
em_mog_update(struct em_mog_context *ctx, double *data)
{
	return_if_fail(ctx != NULL);
	return_if_fail(data != NULL);
	return_if_fail(ctx->means != NULL);

	double a1, a2, *m, *w, *cmx, *vector;
	int i, j, dim, cmx_size, n;

	dim      = ctx->dim;
	cmx_size = dim * dim;
	n        = ctx->ngauss;

	m   = ctx->means;
	w   = ctx->weight;
	cmx = ctx->cov_matrix;

	vector   = malloc(dim * sizeof(double));

	for (i = 0; i < n; i++) {
		a1 = normal_distribution(m, data, cmx, dim);
		a2 = 1 - a1;
		if (a2 == 1.0)
			continue;
		for (j = 0; j < dim; j++) {
			m[j]      = a2 * m[j] + a1 * data[j];
			vector[j] = m[j] - data[j];
		}

		/* TODO: матрица становится неположительно определённой*/
		/*p1 = vector;
		for (j = 0; j < dim; j++) {
			p2 = vector;
			for (l = 0; l < dim; l++) {
					if (i == j)
						*cmx = *cmx * a2 + a1 * (*p1) * (*p2);
				p2++;
				cmx++;
			}
			p1++;
		}*/
		*w = *w * a2 + a1;
		m += dim;
		w++;
	}
	free(vector);
}

void em_mog_reset(struct em_mog_context *ctx, const double *data, int data_size)
{
	double *w, *m, *cov,  weight;
	int i, j, k, dim, cov_size, r;
	k        = ctx->ngauss;
	dim      = ctx->dim;
	cov_size = dim * dim;

	w   = ctx->weight;
	m   = ctx->means;
	cov = ctx->cov_matrix;

	weight = 1.0 / k;

	for (i = 0; i < k; i++) {
		*w = weight;
		memset(cov, 0, cov_size * sizeof(double));
		for (j = 0; j < dim; i++) {
			cov[j + j * dim] = 1.0;
		}
		if (data != NULL && data_size > 0) {
			r = rand() % data_size;
			memcpy(m, data + r, dim * sizeof(double));
		} else {
			memset(m, 0, dim * sizeof(double));
		}
		w++;
		m += 3;
		cov += cov_size;
	}
}
